import os
from tempfile import mktemp

import ants
import ants.utils as utils
import numpy as np
from ants import resample_image_to_target, registration, apply_transforms
from ants.core import ants_image_io as iio
from tqdm import tqdm


def create_template(images: list, masks: list, brain_names: list, atlas_volume: ants.ANTsImage,
                    template_resolution: int = 10, iterations: int = 3, init_with_template=True,
                    save_pre_reg: bool = False, remove_temp_output: bool = False,
                    save_templating_progress: bool = False, pre_registration_type: str = "Rigid",
                    templating_registration_type: str = "SyN") -> ants.ANTsImage:
    """
    Create a template from a folder of images.

    :param images: List of images to use to create the template.
    :type images: list
    :param masks: List of masks to use to create the template.
    :type masks: list
    :param brain_names: List of brain names to use for saving the pre-registered images.
    :type brain_names: list
    :param atlas_volume: Default template to pre-register the brains to and possible the initial volume for registration.
    :type atlas_volume: ants.ANTsImage
    :param template_resolution: The resolution of the template.
    :type template_resolution: int
    :param iterations: The number of iterations to use to create the template.
    :type iterations: int
    :param init_with_template: Whether to initialize the template with the atlas volume or the first image.
    :type init_with_template: bool
    :param save_pre_reg: Whether to save the pre-registered images.
    :type save_pre_reg: bool
    :param remove_temp_output: Whether to remove the temporary output.
    :type remove_temp_output: bool
    :param save_templating_progress: Whether to save the template at each iteration.
    :type save_templating_progress: bool
    :param pre_registration_type: The type of pre-registration to use.
    :type pre_registration_type: str
    :param templating_registration_type: The type of registration to use to create the template.
    :type templating_registration_type: str
    :return: The newly created template.
    :rtype: ants.ANTsImage
    """
    template_images = []
    template_masks = []
    for i, image in tqdm(enumerate(images), desc="Pre-registering images", leave=False, total=len(images), unit="image",
                         position=1):
        image_resampled = ants.resample_image(image, (template_resolution, template_resolution, template_resolution),
                                              use_voxels=False, interp_type=1)
        mask_resampled = ants.resample_image(masks[i], (template_resolution, template_resolution, template_resolution),
                                             use_voxels=False, interp_type=1)

        image_reg, mask_reg = pre_register_brain(image_resampled, mask_resampled, atlas_volume, brain_names[i],
                                                 save_pre_reg=save_pre_reg, registration_type=pre_registration_type)
        template_images.append(image_reg)
        template_masks.append(mask_reg)

    print("Creating template...")
    if init_with_template:
        template = build_template(atlas_volume, template_images, masks=template_masks, iterations=iterations,
                                  save_progress=save_templating_progress, remove_temp_output=remove_temp_output,
                                  type_of_transform=templating_registration_type)
    else:
        template = build_template(template_images[0], template_images, masks=template_masks, iterations=iterations,
                                  save_progress=save_templating_progress, remove_temp_output=remove_temp_output,
                                  type_of_transform=templating_registration_type)
    return template


def pre_register_brain(volume: ants.ANTsImage, mask: ants.ANTsImage, template: ants.ANTsImage, brain: str,
                       save_pre_reg: bool = False, registration_type: str = "Rigid") -> (
        ants.ANTsImage, ants.ANTsImage):
    """
    Register an image to a template and return the registered image and mask.

    :param volume: The volume to register
    :type volume: ants.ANTsImage
    :param mask: The mask to use in registration
    :type mask: ants.ANTsImage
    :param template: The template to register to
    :type template: ants.ANTsImage
    :param brain: The name of the brain
    :type brain: str
    :param save_pre_reg: Whether to save the pre-registered image and mask
    :type save_pre_reg: bool
    :param registration_type: The type of registration to use
    :type registration_type: str
    :return: The registered image and registered mask
    :rtype: tuple[ants.ANTsImage, ants.ANTsImage]
    """
    image_reg_transform = ants.registration(fixed=template, moving=volume, moving_mask=mask,
                                            type_of_transform=registration_type)
    image_reg = ants.apply_transforms(fixed=template, moving=volume, transformlist=image_reg_transform['fwdtransforms'])
    mask_reg = ants.apply_transforms(fixed=template, moving=mask, transformlist=image_reg_transform['fwdtransforms'])
    if save_pre_reg:
        if not os.path.exists("pre_registered"):
            os.makedirs("pre_registered")
        ants.image_write(image_reg, f"pre_registered/{brain}_pre_reg.nii.gz")
        ants.image_write(mask_reg, f"pre_registered/{brain}_pre_reg_mask.nii.gz")
    return image_reg, mask_reg


def build_template(
        initial_template: ants.ANTsImage = None,
        image_list: list[ants.ANTsImage] = None,
        iterations: int = 3,
        gradient_step: float = 0.2,
        blending_weight: float = 0.75,
        weights: bool = None,
        masks: bool = None,
        remove_temp_output: bool = False,
        save_progress: bool = False,
        type_of_transform: str = "SyN",
        **kwargs
) -> ants.ANTsImage:
    """
    Estimate an optimal template from an input image_list
    A modification of the ANTsPy function build_template to use masks.
    Source here: https://antspyx.readthedocs.io/en/latest/_modules/ants/registration/build_template.html#build_template

    :param initial_template: The initial template to use
    :type initial_template: ants.ANTsImage
    :param image_list: The list of images to use to create the template
    :type image_list: list[ants.ANTsImage]
    :param iterations: The number of iterations to use to create the template
    :type iterations: int
    :param gradient_step: For shape update gradient
    :type gradient_step: float
    :param blending_weight: Weight for image blending
    :type blending_weight: float
    :param weights: Weight for each input image
    :type weights: List[float]
    :param masks: List of masks corresponding to the images in image_list
    :type masks: List[ants.ANTsImage]
    :param remove_temp_output: Whether to remove the temporary output files
    :type remove_temp_output: bool
    :param save_progress: Whether to save the progress of the template building
    :type save_progress: bool
    :param type_of_transform: The type of transform to use for registration
    :type type_of_transform: str
    :param kwargs: Extra arguments passed to ants registration
    :return: The newly created template
    :rtype: ants.ANTsImage

    Example
    ^^^^^^^
    >>> import ants
    >>> image = ants.image_read( ants.get_ants_data('r16') )
    >>> image2 = ants.image_read( ants.get_ants_data('r27') )
    >>> image3 = ants.image_read( ants.get_ants_data('r85') )
    >>> timage = ants.build_template( image_list = ( image, image2, image3 ) ).resample_image( (45,45))
    >>> timagew = ants.build_template( image_list = ( image, image2, image3 ), weights = (5,1,1) )
    """

    if weights is None:
        weights = np.repeat(1.0 / len(image_list), len(image_list))
    weights = [x / sum(weights) for x in weights]
    if initial_template is None:
        initial_template = image_list[0] * 0
        for i in range(len(image_list)):
            temp = image_list[i] * weights[i]
            temp = resample_image_to_target(temp, initial_template)
            initial_template = initial_template + temp

    if not os.path.exists("template_progress") and save_progress:
        os.makedirs("template_progress")

    xavg = initial_template.clone()
    for i in tqdm(range(iterations), desc="Running template iterations", leave=False, total=iterations,
                  unit="iteration", position=1):
        for k in range(len(image_list)):
            if masks is None:
                w1 = registration(
                    xavg, image_list[k], type_of_transform=type_of_transform, **kwargs
                )
            else:
                w1 = registration(
                    xavg, image_list[k], type_of_transform=type_of_transform, moving_mask=masks[k], **kwargs
                )
            if k == 0:
                wavg = iio.image_read(w1["fwdtransforms"][0]) * weights[k]
                xavgNew = w1["warpedmovout"] * weights[k]
            else:
                # Remove the transform and warping when not needed, when not last i
                wavg = wavg + iio.image_read(w1["fwdtransforms"][0]) * weights[k]
                xavgNew = xavgNew + w1["warpedmovout"] * weights[k]
                if i < iterations - 1 and remove_temp_output:
                    for fwd_transform in w1["fwdtransforms"]:
                        os.remove(fwd_transform)
                    for inv_transform in w1["invtransforms"]:
                        os.remove(inv_transform)
        print(wavg.abs().mean())
        wscl = (-1.0) * gradient_step
        wavg = wavg * wscl
        wavgfn = mktemp(suffix=".nii.gz")
        iio.image_write(wavg, wavgfn)
        # Keep for debugging/visualization
        xavg = apply_transforms(xavgNew, xavgNew, wavgfn)
        if blending_weight is not None:
            xavg = xavg * blending_weight + utils.iMath(xavg, "Sharpen") * (
                    1.0 - blending_weight
            )
        if save_progress:
            ants.image_write(xavg, f"template_progress/template_{i}.nii.gz")
    return xavg



import math
import os

import PIL.Image
import numpy as np
import pandas as pd
import scipy.ndimage as ndi
from scipy.ndimage import distance_transform_edt
from skimage import measure
from skimage.color import gray2rgb
from skimage.draw import circle_perimeter
from skimage.io import imsave
from skimage.measure import label
from skimage.measure._regionprops import RegionProperties
from skimage.morphology import skeletonize
from skimage.util import img_as_ubyte, img_as_uint
from tqdm import tqdm

PIL.Image.MAX_IMAGE_PIXELS = None


def compute_slice_metrics(output_dir: str, image: np.ndarray, mask: np.ndarray, vessel_mask: np.ndarray,
                          region_map: np.ndarray, vessel_exclude: np.ndarray, voxel_size: float = 0.65) -> None:
    """
    Compute the metrics for a brain slice. Save the results to disk.

    :param output_dir: The directory to save the output to
    :type output_dir: str
    :param image: The image of the brain slice
    :type image: np.ndarray
    :param mask: The mask of the tissue in the brain slice
    :type mask: np.ndarray
    :param vessel_mask: The mask of the vessels in the brain slice
    :type vessel_mask: np.ndarray
    :param region_map: The map of the regions in the brain slice
    :type region_map: np.ndarray
    :param vessel_exclude: The mask of the vessels to exclude from the analysis
    :type vessel_exclude: np.ndarray
    :param voxel_size: The size of the voxels in the image
    :type voxel_size: float
    """

    # Setup output
    if not os.path.exists(output_dir):
        os.mkdir(output_dir)
    df = pd.DataFrame(columns=['image', 'region', 'vessel area (um2)', 'tissue area (um2)', 'vessel density (um2/um2)',
                               'branching points', 'mean diameter (um)'])

    # Get the different brain regions
    regions, region_count = label(region_map, return_num=True)
    props_list = measure.regionprops(regions)

    full_vessel_mask = vessel_mask * mask
    full_vessel_mask = full_vessel_mask * vessel_exclude

    # Compute metrics per region
    for i in tqdm(range(0, region_count), desc="Computing metrics per region for " + image, leave=False):
        region = get_vessel_region(regions, i, full_vessel_mask)

        # Calculate vessel density
        vessel_area, total_area, density = calculate_regional_density(region, i, props_list, output_dir, voxel_size)

        # Count branching points
        branching_points_count, skeleton, branching_points = get_branching_point_count(region, output_dir, filename=str(
            i) + '_skeleton.tif')
        draw_branch_point_circles(skeleton, branching_points, output_dir, filename=str(i) + '_skeleton_circled.png')

        # Calculate average diameter
        mean_diameter = compute_average_diameter(region, skeleton, voxel_size)

        # Save data
        entry = pd.DataFrame.from_dict({'image': [image], 'region': [i], 'vessel area (um2)': [vessel_area],
                                        'tissue area (um2)': [total_area], 'vessel density (um2/um2)': [density],
                                        'branching points': [branching_points_count],
                                        'mean diameter (um)': [mean_diameter]})
        df = pd.concat([df, entry])

    # Compute metrics for the whole slice
    tissue_area, vessel_area, vessel_density = calculate_density(vessel_mask, mask, voxel_size)
    branching_points_count, skeleton, branching_points = get_branching_point_count(vessel_mask, output_dir)
    draw_branch_point_circles(skeleton, branching_points, output_dir)
    mean_diameter = compute_average_diameter(vessel_mask, skeleton, voxel_size)

    # Save intermediate results
    imsave(output_dir + 'regions.png', img_as_ubyte(regions), check_contrast=False)
    imsave(output_dir + 'vessel_exclude.png', img_as_ubyte(vessel_exclude), check_contrast=False)
    imsave(output_dir + '_complete_mask.png', img_as_ubyte(mask), check_contrast=False)
    imsave(output_dir + 'vessels.png', img_as_ubyte(vessel_mask), check_contrast=False)

    # Save data
    entry = pd.DataFrame.from_dict({'image': [image], 'region': 'total', 'vessel area (um2)': [vessel_area],
                                    'tissue area (um2)': [tissue_area], 'vessel density (um2/um2)': [vessel_density],
                                    'branching points': [branching_points_count],
                                    'mean diameter (um)': [mean_diameter]})
    df = pd.concat([df, entry])
    df.to_excel(output_dir + 'regions.xlsx', index=False)


def get_vessel_region(regions: np.ndarray, region_index: int, vessel_mask: np.ndarray) -> np.ndarray:
    """
    Get the vessels in a region.

    :param regions: The regions of the tissue mask.
    :type regions: np.ndarray
    :param region_index: The index of the region.
    :type region_index: int
    :param vessel_mask: The mask of the vessels.
    :type vessel_mask: np.ndarray
    :return: The vessel within the masked region.
    :rtype: np.ndarray
    """
    region = regions == region_index + 1
    region = region * vessel_mask
    return region


def calculate_regional_density(region: np.ndarray, region_index: int, props_list: list[RegionProperties],
                               output_dir: str, voxel_size: float = 0.65) -> tuple[float, float, float]:
    """
    Calculates the density of vessels in a region

    :param region: The region to calculate the density of.
    :type region: np.ndarray
    :param region_index: The computational index of the region.
    :type region_index: int
    :param props_list: The list of properties of the regions.
    :type props_list: list[RegionProperties]
    :param output_dir: The directory to save the region mask to.
    :type output_dir: str
    :param voxel_size: The size of the voxels in the image.
    :type voxel_size: float
    :return: The area of the vessels, the area of the region, and the density of the vessels in a specific region.
    :rtype: tuple[float, float, float]
    """
    vessel_area = (region == 1).sum() * math.pow(voxel_size, 2)
    total_area = props_list[region_index].area * math.pow(voxel_size, 2)
    imsave(output_dir + str(region_index) + '.tif', img_as_ubyte(region), check_contrast=False)
    density = vessel_area / total_area
    return vessel_area, total_area, density


def calculate_density(vessel_mask: np.ndarray, mask: np.ndarray, voxel_size: float = 0.65) -> tuple[
    float, float, float]:
    """
    Calculates the areas of the tissue and vessel to compute the density of vessels in a mask.

    :param vessel_mask: The mask of the vessels.
    :type vessel_mask: np.ndarray
    :param mask: The mask of the tissue.
    :type mask: np.ndarray
    :param voxel_size: The size of the voxels in the image.
    :type voxel_size: float
    :return: The area of the tissue, the area of the vessels, and the density of the vessels.
    :rtype: tuple[float, float, float]
    """
    tissue_area = mask.sum() * math.pow(voxel_size, 2)
    vessel_area = vessel_mask.sum() * math.pow(voxel_size, 2)
    vessel_density = vessel_area / tissue_area
    return tissue_area, vessel_area, vessel_density


def get_branching_point_count(vessel_mask: np.ndarray, output_dir: str, filename: str = 'skeleton.tif') -> tuple[
    int, np.ndarray, np.ndarray]:
    """
    Get the number of branching points in a vessel mask.

    :param vessel_mask: The mask of the vessels.
    :type vessel_mask: np.ndarray
    :param output_dir: The directory to save the skeleton to.
    :type output_dir: str
    :param filename: The filename to save the skeleton to.
    :type filename: str
    :return: The number of branching points in the vessel mask, the skeleton of the vessel mask,
            and the location of the branching points.
    :rtype: tuple[int, np.ndarray, np.ndarray]
    """
    skeleton = skeletonize(vessel_mask)
    branching_points = get_branching_points(skeleton)
    points_count = branching_points.sum()
    imsave(output_dir + filename, img_as_ubyte(skeleton), check_contrast=False)
    return points_count, skeleton, branching_points


def get_branching_points(skeleton: np.ndarray) -> np.ndarray:
    """
    Get the branching points in a skeleton using predefined structural elements
    Source: https://stackoverflow.com/questions/43037692/how-to-find-branch-point-from-binary-skeletonize-image

    :param skeleton: The skeleton of the vessels.
    :type skeleton: np.ndarray
    :return: The branching points in the skeleton.
    :rtype: np.ndarray
    """
    # Setup structural elements for detecting branching points
    selems = list()
    selems.append(np.array([[0, 1, 0], [1, 1, 1], [0, 0, 0]]))
    selems.append(np.array([[1, 0, 1], [0, 1, 0], [1, 0, 0]]))
    selems.append(np.array([[1, 0, 1], [0, 1, 0], [0, 1, 0]]))
    selems.append(np.array([[0, 1, 0], [1, 1, 0], [0, 0, 1]]))
    selems.append(np.array([[0, 0, 1], [1, 1, 1], [0, 1, 0]]))
    selems = [np.rot90(selems[i], k=j) for i in range(5) for j in range(4)]

    # Detect branching points
    branches = np.zeros_like(skeleton, dtype=bool)
    for selem in selems:
        branches |= ndi.binary_hit_or_miss(skeleton, selem)
    return branches


def draw_branch_point_circles(skeleton: np.ndarray, branching_points: np.ndarray, output_dir: str,
                              filename: str = 'skeleton_circled.png') -> None:
    """
    Draw circles around the branching points in a skeleton

    :param skeleton: The skeleton of the vessels.
    :type skeleton: np.ndarray
    :param branching_points: The location of the branching points.
    :type branching_points: np.ndarray
    :param output_dir: The directory to save the skeleton to.
    :type output_dir: str
    :param filename: The filename to save the skeleton to.
    :type filename: str
    :return: The circled branching point in the skeleton.
    :rtype: np.ndarray
    """
    circled_skeleton = gray2rgb(img_as_ubyte(skeleton))
    points_to_draw = np.argwhere(branching_points)
    for point in points_to_draw:
        circy, circx = circle_perimeter(point[0], point[1], 7,
                                        shape=skeleton.shape)
        circled_skeleton[circy, circx] = (220, 20, 20)

    imsave(output_dir + filename, circled_skeleton, check_contrast=False)
    del circled_skeleton


def compute_average_diameter(mask: np.ndarray, skeleton: np.ndarray, voxel_size: float = 0.65) -> float:
    """
    Compute the average diameter of the vessels in a mask

    :param mask: The vessel mask.
    :type mask: np.ndarray
    :param skeleton: The skeleton of the vessels.
    :type skeleton: np.ndarray
    :param voxel_size: The size of the voxels in the image.
    :type voxel_size: float
    :return: The average diameter of the vessels in the mask.
    :rtype: float
    """
    distance = distance_transform_edt(mask)
    radii = distance * skeleton.astype(bool)
    mean_radius = np.mean(radii[radii > 0])
    mean_diameter = 2 * mean_radius
    return mean_diameter * voxel_size


def create_heatmap(image: np.ndarray, output_dir: str, square_size: int = 150) -> None:
    """
    Create and save a heatmap of the vessel density in a brain slice and save it to disk.

    :param image: The image of the brain slice.
    :type image: np.ndarray
    :param output_dir: The directory to save the heatmap to.
    :type output_dir: str
    :param square_size: The size of the squares in the heatmap
    :type square_size: int
    """
    if not os.path.exists(output_dir):
        os.mkdir(output_dir)

    image = img_as_ubyte(image)
    image = image / 255
    image = image.astype(np.uint8)
    heatmap = np.zeros_like(image, dtype=np.uint32)
    x_start = 0
    y_start = 0
    for _ in range(0, int(image.shape[0] / square_size)):
        for j in range(0, int(image.shape[1] / square_size)):
            heatmap[x_start:x_start + square_size, y_start:y_start + square_size] = image[x_start:x_start + square_size,
                                                                                    y_start:y_start + square_size].sum()
            y_start += square_size
        x_start += square_size
        y_start = 0

    # Set final square to max value to ensure same scaling across heatmaps
    heatmap[-1, -1] = (square_size ** 2)
    heatmap = img_as_uint(heatmap)
    heatmap = heatmap.astype(float)
    heatmap = heatmap / (square_size ** 2)
    imsave(output_dir + 'heatmap.tif', heatmap, check_contrast=False)



import ants
from tqdm import tqdm

from liom_toolkit.registration import create_template
from liom_toolkit.segmentation.volume_segmentation import segment_3d_brain


def load_allen_template(resolution, resolution_mm):
    # Load Allen atlas template
    atlas_file = f"<<PATH TO ALLEN TEMPLATE>>"
    atlas_volume = ants.image_read(atlas_file)
    atlas_volume.set_direction([[0., 0., 1.], [1., 0., 0.], [0., -1., 0.]])
    atlas_volume.set_spacing([resolution_mm, resolution_mm, resolution_mm])
    atlas_volume = ants.reorient_image2(atlas_volume, "RAS")
    return atlas_volume


def load_brain(brain, resolution_mm, mirrored=False):
    oct_volume = ants.image_read(f"oct_data/{brain}.nii")
    oct_volume.set_spacing([resolution_mm, resolution_mm, resolution_mm])
    if mirrored:
        oct_volume.set_direction([[-1., 0., 0.], [0., 0., -1.], [0., -1., 0.]])
    else:
        oct_volume.set_direction([[1., 0., 0.], [0., 0., -1.], [0., -1., 0.]])
    oct_volume = ants.reorient_image2(oct_volume, "RAS")
    return oct_volume


def load_brains(brains, resolution_mm):
    oct_volumes = []
    for brain in brains:
        oct_volume = load_brain(brain, resolution_mm)
        oct_volumes.append(oct_volume)

        # Add mirrored volume
        oct_volume = load_brain(brain, resolution_mm, mirrored=True)
        oct_volumes.append(oct_volume)
    return oct_volumes


def load_ras_brain(brain, resolution_mm, mirrored=False):
    oct_volume = ants.image_read(f"oct_data/{brain}.nii")
    oct_volume.set_spacing([resolution_mm, resolution_mm, resolution_mm])
    if mirrored:
        oct_volume.set_direction([[-1., 0., 0.], [0., 1., 0.], [0., 0., 1.]])
    oct_volume = ants.reorient_image2(oct_volume, "RAS")

    return oct_volume


def load_ras_brains(brains, resolution_mm):
    oct_volumes = []
    for brain in brains:
        oct_volume = load_ras_brain(brain, resolution_mm)
        oct_volumes.append(oct_volume)

        # Add mirrored volume
        oct_volume = load_ras_brain(brain, resolution_mm, mirrored=True)
        oct_volumes.append(oct_volume)
    return oct_volumes


def create_mask(volume):
    oct_mask = segment_3d_brain(volume)
    oct_mask = oct_mask.astype("uint8")
    oct_mask = ants.from_numpy(oct_mask)
    oct_mask.set_direction(volume.direction)
    oct_mask.set_spacing(volume.spacing)
    oct_mask.set_origin(volume.origin)
    return oct_mask


def create_masks(volumes):
    oct_masks = []
    for volume in tqdm(volumes, desc="Creating masks", total=len(volumes), unit="volumes"):
        oct_mask = create_mask(volume)
        oct_masks.append(oct_mask)

    return oct_masks


def build_template(brains, ras_brains, resolution_mm, iterations):
    # Register the base to the atlas template

    oct_volumes = load_brains(brains, resolution_mm)

    # Load OCT volumes

    ras_volumes = load_ras_brains(ras_brains, resolution_mm)
    oct_volumes.extend(ras_volumes)

    # Create masks for all volumes
    oct_masks = create_masks(oct_volumes)

    # Apply masks
    oct_volumes = [oct_volume * oct_mask for oct_volume, oct_mask in zip(oct_volumes, oct_masks)]

    # Create template
    template = create_template(oct_volumes, oct_masks, atlas_volume, resolution_mm, iterations)
    return template


# For brains in RSP direction
brains = [

]
# For brains in RAS direction
ras_brains = [

]
resolution = 25
resolution_mm = resolution / 1000
iterations = 15

atlas_volume = load_allen_template(resolution, resolution_mm)
template = build_template(brains, ras_brains, resolution, iterations)

# Register template to atlas
template_transform = ants.registration(fixed=atlas_volume, moving=template, type_of_transform="SyN")
template = ants.apply_transforms(fixed=atlas_volume, moving=template, transformlist=template_transform["fwdtransforms"])
ants.image_write(template, f"templates/oct_average_template_{resolution}um.nii")
